"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.startServer = void 0;
const theme_check_common_1 = require("@shopify/theme-check-common");
const vscode_languageserver_1 = require("vscode-languageserver");
const ClientCapabilities_1 = require("../ClientCapabilities");
const codeActions_1 = require("../codeActions");
const commands_1 = require("../commands");
const completions_1 = require("../completions");
const diagnostics_1 = require("../diagnostics");
const DocumentHighlightsProvider_1 = require("../documentHighlights/DocumentHighlightsProvider");
const documentLinks_1 = require("../documentLinks");
const documents_1 = require("../documents");
const formatting_1 = require("../formatting");
const hover_1 = require("../hover");
const JSONLanguageService_1 = require("../json/JSONLanguageService");
const LinkedEditingRangesProvider_1 = require("../linkedEditingRanges/LinkedEditingRangesProvider");
const RenameProvider_1 = require("../rename/RenameProvider");
const RenameHandler_1 = require("../renamed/RenameHandler");
const utils_1 = require("../utils");
const uri_1 = require("../utils/uri");
const version_1 = require("../version");
const CachedFileSystem_1 = require("./CachedFileSystem");
const Configuration_1 = require("./Configuration");
const defaultLogger = () => { };
/**
 * The `git:` VFS does not support the `fs.readDirectory` call and makes most things break.
 * `git` URIs are the ones you'd encounter when doing a git diff in VS Code. They're not
 * real files, they're just a way to represent changes in a git repository. As such, I don't
 * think we want to sync those in our document manager or try to offer document links, etc.
 *
 * A middleware would be nice but it'd be a bit of a pain to implement.
 */
const hasUnsupportedDocument = (params) => {
    return ('textDocument' in params &&
        'uri' in params.textDocument &&
        typeof params.textDocument.uri === 'string' &&
        params.textDocument.uri.startsWith('git:'));
};
/**
 * This code runs in node and the browser, it can't talk to the file system
 * or make requests. Stuff like that should be injected.
 *
 * In browser, theme-check-js wants these things:
 *   - fileExists(path)
 *   - defaultTranslations
 *
 * Which means we gotta provide 'em from here too!
 */
function startServer(connection, { fs: injectedFs, loadConfig, log = defaultLogger, jsonValidationSet, themeDocset: remoteThemeDocset, }) {
    const fs = new CachedFileSystem_1.CachedFileSystem(injectedFs);
    const fileExists = (0, theme_check_common_1.makeFileExists)(fs);
    const clientCapabilities = new ClientCapabilities_1.ClientCapabilities();
    const configuration = new Configuration_1.Configuration(connection, clientCapabilities);
    const documentManager = new documents_1.DocumentManager(fs);
    const diagnosticsManager = new diagnostics_1.DiagnosticsManager(connection);
    const documentLinksProvider = new documentLinks_1.DocumentLinksProvider(documentManager, findThemeRootURI);
    const codeActionsProvider = new codeActions_1.CodeActionsProvider(documentManager, diagnosticsManager);
    const onTypeFormattingProvider = new formatting_1.OnTypeFormattingProvider(documentManager, async function setCursorPosition(textDocument, position) {
        if (!clientCapabilities.hasShowDocumentSupport)
            return;
        connection.sendRequest(vscode_languageserver_1.ShowDocumentRequest.type, {
            uri: textDocument.uri,
            takeFocus: true,
            selection: {
                start: position,
                end: position,
            },
        });
    });
    const linkedEditingRangesProvider = new LinkedEditingRangesProvider_1.LinkedEditingRangesProvider(documentManager);
    const documentHighlightProvider = new DocumentHighlightsProvider_1.DocumentHighlightsProvider(documentManager);
    const renameProvider = new RenameProvider_1.RenameProvider(documentManager);
    const renameHandler = new RenameHandler_1.RenameHandler(connection, clientCapabilities, documentManager, fileExists);
    async function findThemeRootURI(uri) {
        const rootUri = await (0, theme_check_common_1.findRoot)(uri, fileExists);
        const config = await loadConfig(rootUri, fs);
        return config.rootUri;
    }
    const getMetafieldDefinitionsForRootUri = (0, theme_check_common_1.memoize)((0, theme_check_common_1.makeGetMetafieldDefinitions)(fs), (rootUri) => rootUri);
    const getMetafieldDefinitions = async (uri) => {
        const rootUri = await findThemeRootURI(uri);
        return getMetafieldDefinitionsForRootUri(rootUri);
    };
    // These are augmented here so that the caching is maintained over different runs.
    const themeDocset = new theme_check_common_1.AugmentedThemeDocset(remoteThemeDocset);
    const runChecks = (0, utils_1.debounce)((0, diagnostics_1.makeRunChecks)(documentManager, diagnosticsManager, {
        fs,
        loadConfig,
        themeDocset,
        jsonValidationSet,
        getMetafieldDefinitions,
    }), 100);
    const getTranslationsForURI = async (uri) => {
        const rootURI = await findThemeRootURI(uri);
        const theme = documentManager.theme(rootURI);
        const getDefaultTranslations = (0, theme_check_common_1.makeGetDefaultTranslations)(fs, theme, rootURI);
        const [defaultTranslations, shopifyTranslations] = await Promise.all([
            getDefaultTranslations(),
            themeDocset.systemTranslations(),
        ]);
        return { ...shopifyTranslations, ...defaultTranslations };
    };
    const getSchemaTranslationsForURI = async (uri) => {
        const rootURI = await findThemeRootURI(uri);
        const theme = documentManager.theme(rootURI);
        const getDefaultSchemaTranslations = (0, theme_check_common_1.makeGetDefaultSchemaTranslations)(fs, theme, rootURI);
        return getDefaultSchemaTranslations();
    };
    const snippetFilter = ([uri]) => /\.liquid$/.test(uri) && /snippets/.test(uri);
    const getSnippetNamesForURI = async (uri) => {
        const rootUri = await findThemeRootURI(uri);
        const snippetUris = await (0, theme_check_common_1.recursiveReadDirectory)(fs, rootUri, snippetFilter);
        return snippetUris.map(uri_1.snippetName);
    };
    const getThemeSettingsSchemaForURI = async (uri) => {
        try {
            const rootUri = await findThemeRootURI(uri);
            const settingsSchemaUri = theme_check_common_1.path.join(rootUri, 'config', 'settings_schema.json');
            const contents = await fs.readFile(settingsSchemaUri);
            const json = (0, theme_check_common_1.parseJSON)(contents);
            if ((0, theme_check_common_1.isError)(json) || !Array.isArray(json)) {
                throw new Error('Settings JSON file not in correct format');
            }
            return json;
        }
        catch (error) {
            console.error(error);
            return [];
        }
    };
    const getModeForURI = async (uri) => {
        const rootUri = await (0, theme_check_common_1.findRoot)(uri, fileExists);
        const config = await loadConfig(rootUri, fs);
        return config.context;
    };
    const jsonLanguageService = new JSONLanguageService_1.JSONLanguageService(documentManager, jsonValidationSet, getSchemaTranslationsForURI, getModeForURI);
    const completionsProvider = new completions_1.CompletionsProvider({
        documentManager,
        themeDocset,
        getTranslationsForURI,
        getSnippetNamesForURI,
        getThemeSettingsSchemaForURI,
        log,
        getMetafieldDefinitions,
    });
    const hoverProvider = new hover_1.HoverProvider(documentManager, themeDocset, getMetafieldDefinitions, getTranslationsForURI, getThemeSettingsSchemaForURI);
    const executeCommandProvider = new commands_1.ExecuteCommandProvider(documentManager, diagnosticsManager, clientCapabilities, runChecks, connection);
    connection.onInitialize((params) => {
        clientCapabilities.setup(params.capabilities, params.initializationOptions);
        jsonLanguageService.setup(params.capabilities);
        configuration.setup();
        const fileOperationRegistrationOptions = {
            filters: [
                {
                    pattern: {
                        glob: '**/*.{liquid,json}',
                    },
                },
                {
                    pattern: {
                        glob: '**/assets/*',
                    },
                },
            ],
        };
        const result = {
            capabilities: {
                textDocumentSync: {
                    change: vscode_languageserver_1.TextDocumentSyncKind.Full,
                    save: true,
                    openClose: true,
                },
                codeActionProvider: {
                    codeActionKinds: [...codeActions_1.CodeActionKinds],
                },
                completionProvider: {
                    triggerCharacters: ['.', '{{ ', '{% ', '<', '/', '[', '"', "'", ':'],
                },
                documentOnTypeFormattingProvider: {
                    firstTriggerCharacter: ' ',
                    moreTriggerCharacter: ['{', '%', '-', '>'],
                },
                documentLinkProvider: {
                    resolveProvider: false,
                    workDoneProgress: false,
                },
                documentHighlightProvider: true,
                linkedEditingRangeProvider: true,
                renameProvider: {
                    prepareProvider: true,
                },
                executeCommandProvider: {
                    commands: [...commands_1.Commands],
                },
                hoverProvider: {
                    workDoneProgress: false,
                },
                workspace: {
                    fileOperations: {
                        didRename: fileOperationRegistrationOptions,
                        didCreate: fileOperationRegistrationOptions,
                        didDelete: fileOperationRegistrationOptions,
                    },
                },
            },
            serverInfo: {
                name: 'theme-language-server',
                version: version_1.VERSION,
            },
        };
        return result;
    });
    connection.onInitialized(() => {
        log(`[SERVER] Let's roll!`);
        configuration.fetchConfiguration();
        configuration.registerDidChangeCapability();
        configuration.registerDidChangeWatchedFilesNotification({
            watchers: [
                {
                    globPattern: '**/.shopify/*',
                },
            ],
        });
    });
    connection.onDidChangeConfiguration((_params) => {
        configuration.clearCache();
    });
    connection.onDidOpenTextDocument(async (params) => {
        if (hasUnsupportedDocument(params))
            return;
        const { uri, text, version } = params.textDocument;
        documentManager.open(uri, text, version);
        if (await configuration.shouldCheckOnOpen()) {
            runChecks([uri]);
        }
    });
    connection.onDidChangeTextDocument(async (params) => {
        if (hasUnsupportedDocument(params))
            return;
        const { uri, version } = params.textDocument;
        documentManager.change(uri, params.contentChanges[0].text, version);
        if (await configuration.shouldCheckOnChange()) {
            runChecks([uri]);
        }
        else {
            // The diagnostics may be stale! Clear em!
            diagnosticsManager.clear(params.textDocument.uri);
        }
    });
    connection.onDidSaveTextDocument(async (params) => {
        if (hasUnsupportedDocument(params))
            return;
        const { uri } = params.textDocument;
        fs.readFile.invalidate(uri);
        fs.stat.invalidate(uri);
        if (await configuration.shouldCheckOnSave()) {
            runChecks([uri]);
        }
    });
    connection.onDidCloseTextDocument((params) => {
        if (hasUnsupportedDocument(params))
            return;
        const { uri } = params.textDocument;
        documentManager.close(uri);
        diagnosticsManager.clear(uri);
    });
    connection.onDocumentLinks(async (params) => {
        if (hasUnsupportedDocument(params))
            return [];
        return documentLinksProvider.documentLinks(params.textDocument.uri);
    });
    connection.onCodeAction(async (params) => {
        return codeActionsProvider.codeActions(params);
    });
    connection.onExecuteCommand(async (params) => {
        await executeCommandProvider.execute(params);
    });
    connection.onCompletion(async (params) => {
        var _a;
        if (hasUnsupportedDocument(params))
            return [];
        return ((_a = (await jsonLanguageService.completions(params))) !== null && _a !== void 0 ? _a : (await completionsProvider.completions(params)));
    });
    connection.onHover(async (params) => {
        var _a;
        if (hasUnsupportedDocument(params))
            return null;
        return (_a = (await jsonLanguageService.hover(params))) !== null && _a !== void 0 ? _a : (await hoverProvider.hover(params));
    });
    connection.onDocumentOnTypeFormatting(async (params) => {
        if (hasUnsupportedDocument(params))
            return null;
        return onTypeFormattingProvider.onTypeFormatting(params);
    });
    connection.onDocumentHighlight(async (params) => {
        if (hasUnsupportedDocument(params))
            return [];
        return documentHighlightProvider.documentHighlights(params);
    });
    connection.onPrepareRename(async (params) => {
        if (hasUnsupportedDocument(params))
            return null;
        return renameProvider.prepare(params);
    });
    connection.onRenameRequest(async (params) => {
        if (hasUnsupportedDocument(params))
            return null;
        return renameProvider.rename(params);
    });
    connection.languages.onLinkedEditingRange(async (params) => {
        if (hasUnsupportedDocument(params))
            return null;
        return linkedEditingRangesProvider.linkedEditingRanges(params);
    });
    connection.onDidChangeWatchedFiles(async (params) => {
        if (params.changes.length === 0)
            return;
        for (const change of params.changes) {
            fs.readDirectory.invalidate(theme_check_common_1.path.dirname(change.uri));
            fs.readFile.invalidate(change.uri);
            fs.stat.invalidate(change.uri);
            if (change.uri.endsWith('metafields.json')) {
                const rootUri = await findThemeRootURI(change.uri);
                getMetafieldDefinitionsForRootUri.invalidate(rootUri);
            }
        }
    });
    // These notifications could cause a MissingSnippet check to be invalidated
    //
    // It is not guaranteed that the file is or was opened when it was
    // created/renamed/deleted. If we're smart, we're going to re-lint for
    // every root affected. Unfortunately, we can't just use the debounced
    // call because we could run in a weird timing issue where didOpen
    // happens after the didRename and causes the 'lastArgs' to skip over the
    // ones we were after.
    //
    // So we're using runChecks.force for that.
    connection.workspace.onDidCreateFiles((params) => {
        const triggerUris = params.files.map((fileCreate) => fileCreate.uri);
        runChecks.force(triggerUris);
        for (const { uri } of params.files) {
            fs.readDirectory.invalidate(theme_check_common_1.path.dirname(uri));
            fs.readFile.invalidate(uri);
            fs.stat.invalidate(uri);
        }
    });
    connection.workspace.onDidRenameFiles((params) => {
        const triggerUris = params.files.map((fileRename) => fileRename.newUri);
        renameHandler.onDidRenameFiles(params);
        runChecks.force(triggerUris);
        for (const { oldUri, newUri } of params.files) {
            documentManager.delete(oldUri);
            fs.readDirectory.invalidate(theme_check_common_1.path.dirname(oldUri));
            fs.readDirectory.invalidate(theme_check_common_1.path.dirname(newUri));
            fs.readFile.invalidate(oldUri);
            fs.readFile.invalidate(newUri);
            fs.stat.invalidate(oldUri);
            fs.stat.invalidate(newUri);
        }
    });
    connection.workspace.onDidDeleteFiles((params) => {
        const triggerUris = params.files.map((fileDelete) => fileDelete.uri);
        runChecks.force(triggerUris);
        for (const { uri } of params.files) {
            documentManager.delete(uri);
            fs.readDirectory.invalidate(theme_check_common_1.path.dirname(uri));
            fs.readFile.invalidate(uri);
            fs.stat.invalidate(uri);
        }
    });
    connection.listen();
}
exports.startServer = startServer;
//# sourceMappingURL=startServer.js.map